# main file for data visualisation
# analyse and plot different aspects of the LP and layup space

import numpy as np

# matplotlib pgf export settings -> TeX
import matplotlib as mpl

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
plt.style.use('seaborn-white')
import pandas as pd
from lpfuncs import *
from pyDOE import * 
import seaborn as sns

# plot exporting without needing a tex install inside python env 
import tikzplotlib
# latex settings
pgf_with_latex = {
        #"pgf.texsystem": "pdflatex",
        'text.usetex': False,
        'pgf.rcfonts': False,
        "pgf.preamble": [
                r"\usepackage[utf8x]{inputenc}"
                ]
}
#font settings
mpl.rcParams['font.family'] = 'serif'

# define data
dt = 'f4'
layers = 6 # set N
delta = 2 # allow changing layer angle in these increments
angles = 180//delta 
length = angles**layers

num_lp = 12
lpnames = ['$V_1^A$', '$V_2^A$', '$V_3^A$', '$V_4^A$', '$V_1^B$', '$V_2^B$', '$V_3^B$', '$V_4^B$', '$V_1^D$', '$V_2^D$', '$V_3^D$', '$V_4^D$']

# select what to plot
doe = 1
if doe:
    # generate samples instead of using array data
    samples = 100000
    # generate valid laminates via LHS
    y = np.deg2rad(np.rint(lhs(layers,samples)*(angles-1))*delta-(90-delta))
    # calculate LP from combinations
    #print(y.shape)
    X = calc_lp_array(y)
    print(X.shape)
    #df = pd.DataFrame(X.transpose())#, index=index)
    #df.info()
    #print(df.describe())

# select what to plot
hist_doe = 0
anglescatter_doe = 0
corr_doe = 0
corr_mat_doe = 0
andrews_doe = 0
contour_doe = 0
scatter_jac_doe = 0
k_factor = 0
sspace_doe = 0
solutionplot = 0
solution_implicit = 0
solution_shapes = 0
error_plot = 1

if hist_doe:
    # calc histograms of LP
    b = 101#201 # number of bins
    #plt.hist(x, bins=b, weights=ha, density=True)
    #plt.show()
    # create figure with subfigures
    fig = plt.figure()
    fig.subplots_adjust(hspace=0.6, wspace=0.4)
    for i in range(1, num_lp+1):
        ax = fig.add_subplot(3, 4, i) # add sharex option -> subplots!
        ax.hist(X[:, i-1], bins=b, density=True)
        ax.tick_params(axis='both', which='both', length=5)
        ax.set_xlim(-1, 1) 
        ax.set_xticks(np.array([-1, -0.5, 0, 0.5, 1]))
        ax.set_xticklabels(["-1", "", "0", "", "1"])
        ax.set_title(lpnames[i-1])
        #ax[i, lp].set_yticks(np.array([-90, -45, 0, 45, 90])) 
        #ax[i, lp].set_ylim(0,) 
    fig.text(0.5, 0.02, 'LP', ha='center')
    fig.text(0.06, 0.5, 'Density', va='center', rotation='vertical')
    #tikzplotlib.save("plots/"+"histdoe"+str(layers)+"_"+str(samples)+".tex", standalone=True)
    plt.show()

# plot angle dependency of each LP in each layer
if anglescatter_doe:
    #layers = 2
    #mpl.pyplot.figure(figsize=(9, 6))
    mpl.rcParams.update({'font.size': 8})
    # set line width for smaller borders
    mpl.rcParams['axes.linewidth'] = 0.5
    y = np.rad2deg(y) # convert all angles in degree
    fig2, ax = plt.subplots(layers, num_lp, sharex='col', sharey='row', figsize=(10, 8)) #6 for 4 layers
    l = list(range(0, layers)) 
    l_rev = l[::-1] # plot first layer lowest
    for i in l: 
        for lp in range(0, num_lp):
            xpoints = X[:,lp]
            ypoints = y[:,l_rev[i]]#get_comb_array(i, layers, angles, delta)
            #ax = fig2.add_subplot(layers, num_lp, i)
            # print minimum LP at angle 0
            if lp == 0:
                # calc min value
                #atol = 1e-5#np.deg2rad(delta/2)
                #zeros = np.argwhere(np.isclose(ypoints, 0, atol=atol)) # get angles near zero
                #zero_lps = xpoints[zeros]
                #print("Layer", i, "LP", lp, "Min lp at 0°", np.amin(zero_lps))
                # put y label
                ax[i, lp].set_ylabel(str(l_rev[i]+1))
            if i == layers-1:
                ax[i, lp].set_xlabel(lpnames[lp])
                
            ret = ax[i, lp].hist2d(xpoints, ypoints, bins=20, density=True)#scatter(xpoints, ypoints)
            ax[i, lp].tick_params(axis='both', which='both', length=5)
            ax[i, lp].set_xlim(-1, 1) 
            ax[i, lp].set_xticks(np.array([-1, 0, 1])) 
            ax[i, lp].set_yticks(np.array([-90, -45, 0, 45, 90])) 
            ax[i, lp].set_ylim(-90,90) 
            #ax[i, lp].set_frame_on(True)
            #ax[i, lp].hexbin(xpoints, ypoints, bins=20, gridsize=20)#scatter(xpoints, ypoints)
    #print(ret)
    # plot common axis  
    fig2.text(0.5, 0.04, 'LP', ha='center')
    fig2.text(0.05, 0.5, 'angle in ° in ply', va='center', rotation='vertical')
    fig2.savefig("anglescatter_6.pdf", orientation='portrait', pad_inches=0, bbox_inches='tight')
    # export
    #tikzplotlib.save("plots/"+"anglesdoe"+str(layers)+"_"+str(samples)+".tex", standalone=True)
    plt.show()
    #plt.savefig('plots/anglescatter.pgf', format='pgf')
    #plt.savefig("plots/seaborn.pdf")

# yields nice correlation patterns, but otherwise not really useful
if corr_doe:
    fig2, ax = plt.subplots(num_lp, num_lp, sharex='col', sharey='row')
    for i in range(0, num_lp):
        for k in range(0, num_lp):
            ax[i, k].hist2d(X[:,i], X[:,k], bins=50)
    plt.show()

# correlation matrix
if corr_mat_doe:
    df = pd.DataFrame(X)
    corr = df.corr()
    # rename labels
    corr.index = lpnames
    corr.columns = lpnames 
    #print(corr)
    #corr.to_csv("plots/"+"corrheatmap"+str(layers)+"_"+str(samples)+".csv")
    df = corr
    with open("plots/"+"corrheatmap"+str(layers)+"_"+str(samples)+".csv","w") as f:
        f.write("x y z xl yl\n")
        for i,line in enumerate(df.values):
            for j,val in enumerate(line):
                f.write("{} {} {} {} {}\n".format(j,i,val, df.columns[j], df.index[j]))
            f.write("\n")
    ax = sns.heatmap(
    corr,
    annot=True,
    fmt=".2f", #".1g"
    vmin=-1, vmax=1, center=0,
    cmap=sns.diverging_palette(20, 220, n=200),
    square=True
    )
    # tikzplotlib doesn't work here
    plt.show()

# andrew curves, doesn't yield useful info
if andrews_doe:
    andrew = pd.DataFrame({'V1A': X[:,0],
                           'V2A': X[:,1],
                           'V3A': X[:,2],
                           'V4A': X[:,3],
                           'V1B': X[:,4],
                           'V2B': X[:,5],
                           'V3B': X[:,6],
                           'V4B': X[:,7],
                           'V1D': X[:,8],
                           'V2D': X[:,9],
                           'V3D': X[:,10],
                           'V4D': X[:,11]})
    andrew = pd.DataFrame({'1': [30, -50],'2': [-30, 10],'3': [60, 35],'4':[0,35]})
    #andrew = andrew.sample(n=10)
    andrew['LP']='LP' 
    andrew.info()
    #andrew = pd.pivot(andrew)
    andrew.info()
    pd.plotting.andrews_curves(andrew, 'LP')
    plt.show()

if contour_doe:
    # compute gradient info
    jacs = calc_lp_jac_array_doe(y)
    #fig5, ax5 = plt.subplots(4, 3, sharex='col', sharey='row')
    #for i in range(0, layers):
    #    for k in range(0, layers):
    #        ax5[i, k].contour(y[:,i], y[:,k], bins=50)
    #        ax5[i, k].colorbar()
    fig = plt.figure()
    fig.subplots_adjust(hspace=0.4, wspace=0.4)
    for i in range(1, num_lp+1):
        ax = fig.add_subplot(3, 4, i)
        # only for first two layer! (2D plot!)
        # plot gradient of first angle (theta_0)
        clplt = ax.tricontourf(y[:,4], y[:,5], jacs[:, 4, i-1], levels=14, cmap="RdBu_r")
        #ax.colorbar()
        fig.colorbar(clplt, ax=ax)
    plt.show()

# NOT ready yet -> may not be useful at all
if scatter_jac_doe:
    jac = calc_lp_jac_array_doe(y)
    tol = 1e-3 # how close to 0 to count as extremum (remember, these values come from gridsearch!)
    zeros = np.argwhere(np.isclose(jac, 0, atol=tol)) # get elements near zero
    zeros[zeros[:,i]]
    # select those coordinates (angles) where jacobian is 0 
    #overlay scatter plots to obtain different markers
    for i in range(i, num_lp):
        plt.scatter()
    #plt.show()

# plot s factors (fornerly k factors)
if k_factor:
    N = 12
    Z2 = np.array([((-N/2+i+1)**2 -(-N/2+i)**2) for i in range(N)]).astype(np.float32)
    Z3 = np.array([((-N/2+i+1)**3 -(-N/2+i)**3) for i in range(N)]).astype(np.float32)
    N2 = 2/(N**2)
    N3 = 4/(N**3)
    x = np.linspace(1, N, N)
    
    y1 = np.full((N, ), 1/N)
    y2 = Z2*N3
    y3 = Z3*N3
    print(y2)
    print(y3)
    
    plt.xlabel('$N$')
    plt.ylabel('$k$')
    plt.xticks(np.arange(0, N+1, 1)) 
    #print(N2)
    plt.grid(True)
    plt.plot(x,y1, 'o-', label='$k_1$')
    plt.plot(x,y2, '^-', label='$k_2$')
    plt.plot(x,y3, 's-', label='$k_3$')
    plt.legend(loc="upper center")
    # standalone true!
    #tikzplotlib.save("plots/"+"kfactors_N"+str(N)+".tex")
    plt.show()

# plot solution spaces
if sspace_doe:
    # only 3D or 2D so only 2 / 3 layers
    # print one space (e.g. V1A) where the LP is one value (e.g. 0.5)
    # need good resolution -> many samples
    # scatter plot? / line plot would be enough to see cross polytope shape
    # can only display in euclidean space though ....
    # from LP to dostance on unit sphere [-1;1] -> [2;0]
    tlpval = 0.3
    tol = 0.01
    nlp = 0 # V1A
    # modify lp array so only values around tlpval +- tol remain
    modindices_l = np.argwhere(X[:,nlp]>(tlpval-tol)).flatten()
    modindices_u = np.argwhere(X[:,nlp]<(tlpval+tol)).flatten()
    # take only the same indices
    modindices = np.intersect1d(modindices_l, modindices_u)
    #Xmod = X[modindices,nlp]
    ymod = y[modindices] # only angles that are near the specified lp
    # plot these values
    # 1D, fix axis to [-90°,90°] 
    #plt.scatter(ymod[:,0], np.zeros_like(ymod[:,0]))#, ymod[:,1])
    # 2D, fix axis to [-90°,90°] 
    #plt.scatter(ymod[:,0], ymod[:,1])
    # 3D, fix axis to [-90°,90°] 
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(ymod[:,0], ymod[:,1], ymod[:,2])
    #ax.plot_trisurf(ymod[:,0], ymod[:,1], ymod[:,2])
    plt.show()

# plot solution surface for 2D (or 3D) laminate
if solutionplot:
    # plot dist to target lp for 2,3,4 dimensions
    # need to adjust for dimensions separatly
    # generate coordinate grid
    dim = 2
    delta = 5
    #angles = np.deg2rad(np.linspace(64.999,65.001,50))
    angles = np.deg2rad(np.arange(-90+delta,90+delta,delta))
    lams = np.array(np.meshgrid(angles, angles)).T.reshape(-1, dim)
    print(lams.shape)
    lps = calc_lp_array(lams)
    # create target laminate
    lam_t = np.deg2rad(np.array([45, 15])) #70, 10
    lp_t = get_lp(lam_t)
    print(lp_t)
    # transform lams back to deg for better plotting
    lams = np.rad2deg(lams)
    dists = np.sqrt(((lps - lp_t)**2).sum(axis=1))
    # 2D contour
    #plt.contourf(lams[:,0].reshape((36,36)), lams[:,1].reshape((36,36)), dists.reshape((36,36)), levels = 10, cmap=plt.hot())

    # 3D
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.plot_trisurf(lams[:,0], lams[:,1], dists)
    
    # 4D, scatter with color info
    #img = ax.scatter(lams[:,0],lams[:,1],lams[:,2], c=dists, cmap=plt.hot())
    #fig.colorbar(img)
   
    plt.show()
    # export datapoints for plotting
    a =  np.vstack((lams[:,0], lams[:,1], dists))
    data = pd.DataFrame(data=a.T)#, index=["x", "y", "z"])#, INdex=index)
    data.info()
    # save to file
    data.to_csv("plots/solutionplot.csv")

#plot the implicit solution to export to pgf
if solution_implicit:
    # [0.4330127,  0.75     , -0.25     ,  0.4330127,  
    #  0.4330127, -0.25     ,  0.75     ,  0.4330127,
    #  0.4330127,  0.75     , -0.25     ,  0.4330127]
    # [45°,15°] laminate
    #plt.figure(figsize=(4,4))
    #fig, ax = plt.subplots(1,1)
    xx = np.linspace(-np.pi/2,np.pi/2,100)
    yy = xx
    x,y = np.meshgrid(xx,yy)
    xdeg = np.rad2deg(x)
    ydeg = np.rad2deg(y)
    # V1A
    cs = plt.contour(xdeg, ydeg, ((np.cos(2*x)+np.cos(2*y))/2-0.433), levels=[0]);
    plt.setp(cs.collections[0], label=r"$V_1^A$", linestyle="-", color="blue")
    ax = plt.gca()
    # V2A
    cs = plt.contour(xdeg, ydeg, ((np.sin(2*x)+np.sin(2*y))/2-0.75), levels=[0]);
    plt.setp(cs.collections[0], label=r"$V_2^A$", linestyle="-", color="red")
    # V3A
    #cs = plt.contour(xdeg, ydeg, ((np.cos(4*x)+np.cos(4*y))/2+0.25), levels=[0]);
    #plt.setp(cs.collections[0], label=r"$V_3^A$", linestyle="-", color="green")
    # V4A
    #cs = plt.contour(xdeg, ydeg, ((np.sin(4*x)+np.sin(4*y))/2-0.433), levels=[0]);
    #plt.setp(cs.collections[0], label=r"$V_4^A$", linestyle="-", color="black")
    # V1B
    cs = plt.contour(xdeg, ydeg, ((-np.cos(2*x)+np.cos(2*y))/2-0.433), levels=[0]);
    plt.setp(cs.collections[0], label=r"$V_1^B$", linestyle="--", color="blue")
    # V2B
    cs = plt.contour(xdeg, ydeg, ((-np.sin(2*x)+np.sin(2*y))/2+0.25), levels=[0]);
    plt.setp(cs.collections[0], label=r"$V_2^B$", linestyle="--", color="red")
    # V3B
    #cs = plt.contour(xdeg, ydeg, ((-np.cos(4*x)+np.cos(4*y))/2-0.75), levels=[0]);
    #plt.setp(cs.collections[0], label=r"$V_3^B$", linestyle="--", color="green")
    # V4B
    #cs = plt.contour(xdeg, ydeg, ((-np.sin(4*x)+np.sin(4*y))/2-0.433), levels=[0]);
    #plt.setp(cs.collections[0], label=r"$V_4^B$", linestyle="--", color="black")
    #ax.set_yticks(np.array([-90, -45, 0, 45, 90])) 
    #ax.set_xticks(np.array([-90, -45, 0, 45, 90])) 
    #ax.tick_params(axis='both', which='both', length=5)
    #ax.set_xticks(np.array([-90, -45, 0, 45, 90])) 
    
    plt.legend(loc='upper left')
    #ax.set_aspect('equal','datalim')
    plt.xlim(-90,90)
    plt.ylim(-90,90)
    ax.set_aspect('equal')
    #tikzplotlib.save("plots/"+"solution_implicit"+".tex", standalone=True)
    plt.show()

if solution_shapes:
    xx = np.linspace(-np.pi/2,np.pi/2,100)
    yy = xx
    x,y = np.meshgrid(xx,yy)
    xdeg = np.rad2deg(x)
    ydeg = np.rad2deg(y)
    # V1A
    cs = plt.contour(xdeg, ydeg, ((np.cos(2*x)+np.cos(2*y))/2-0), levels=[0]);
    plt.setp(cs.collections[0], label=r"$V_1^A$", linestyle="-", color="blue")
    ax = plt.gca()
    # V1A
    cs = plt.contour(xdeg, ydeg, ((np.cos(2*x)+np.cos(2*y))/2-0.7), levels=[0]);
    plt.setp(cs.collections[0], label=r"$V_1^A$", linestyle="-", color="red")
    ax = plt.gca()
    # V1A
    cs = plt.contour(xdeg, ydeg, ((np.cos(2*x)+np.cos(2*y))/2-0.3), levels=[0]);
    plt.setp(cs.collections[0], label=r"$V_1^A$", linestyle="-", color="green")
    ax = plt.gca()
    # V1A
    cs = plt.contour(xdeg, ydeg, ((np.cos(2*x)+np.cos(2*y))/2+0.3), levels=[0]);
    plt.setp(cs.collections[0], label=r"$V_1^A$", linestyle="-", color="yellow")
    ax = plt.gca()
    
    plt.legend(loc='upper left')
    #ax.set_aspect('equal','datalim')
    plt.xlim(-90,90)
    plt.ylim(-90,90)
    ax.set_aspect('equal')
    #tikzplotlib.save("plots/"+"solution_shapes"+".tex", standalone=True)
    plt.show()

# interpret the error
# how much deviation in the layer angles results in how much error?
if error_plot:
    num = 20 # number of sampling points   
    layer_nums = np.array([6, 12, 20, 60])
    thetas = np.linspace(0, 5, num)
    meanMSEs_m = np.zeros((layer_nums.size, num))
    meanMSEs_u = np.zeros((layer_nums.size, num))
    samples = 1000 # random laminate samples
    for k, layers in enumerate(layer_nums):
        # setting from doe generation at top!
        y = np.deg2rad(np.rint(lhs(layers,samples)*(angles-1))*delta-(90-delta))
        # calculate LP from combinations
        X = calc_lp_array(y)

        for i in range(num):
            # create random start laminate from doe
            t_max = np.deg2rad(thetas[i])# max total deviation in RAD
            # maximum impact error (only one outer layers)
            y_err = y.copy()
            y_err[:,0] += t_max#*layers
            # calc new lps
            lp_err = calc_lp_array(y_err)
            meanMSEs_m[k,i] = np.mean(np.abs(lp_err-X)) #mean of all lp error and all samples -> scalar
            # uniformly distributed error
            y_err = y + (np.random.uniform(size=(y.shape[0], y.shape[1]))*2-1)*t_max
            lp_err = calc_lp_array(y_err)
            meanMSEs_u[k,i] = np.mean((lp_err-X)**2)/layers
        #plt.plot(thetas,meanMSEs_m[k], 'o-', label=layers)
        plt.plot(thetas,meanMSEs_u[k], '^-', label=layers)
    plt.legend()
    plt.show()

